#include "tb.h"

#include <stdlib.h>
#include <memory.h>
#include <string.h>

#include "cpptypes.h"
#include "debug.h"

//          
//   ,       
//   

void tb_reallocate(struct tb_table *_t, int lines)
{
  //      
  const int lines_step=1000;
  int lines_new_count=0;
  int lines_mem_delta;

  if(_t->lines_count_mem>=lines)
    return;

  //lines_new_count=(lines/lines_step+1)*lines_step;
  lines_new_count=_t->lines_count_mem*_t->lines_count_mem+1;

  lines_mem_delta=_t->lines_mem-_t->lines;

  _t->lines_mem=realloc(_t->lines_mem,sizeof(tb_line)*lines_new_count);
  memset(&_t->lines_mem[_t->lines_count_mem],0,(lines_new_count-_t->lines_count_mem)*sizeof(tb_line));

  _t->lines=_t->lines_mem+lines_mem_delta;

  _t->lines_count_mem=lines_new_count;
}

void tb_free_lines(struct tb_table *_t)
{
  int i;
  for(i=0;i<_t->lines_count_mem;i++)
  {
    if(_t->lines_mem[i]!=0)
    {
      free(_t->lines_mem[i]);
      _t->lines_mem[i]=0;
    }
  }
}

void tb_initialize(struct tb_table *_t)
{
  _t->lines=0;
  _t->lines_mem=0;
  _t->lines_count_mem=0;
  _t->lines_count=0;
  _t->columns=0;
}

void tb_describe(struct tb_table *_t, struct tb_column *_col, int _n, int _occurs)
{
  int i,offset=0;

  //         
  //       

  tb_free_lines(_t);
  tb_clear(_t);

  _t->columns_count=_n;
  _t->columns=malloc(sizeof(struct tb_column)*_n);
  memcpy(_t->columns,_col,sizeof(struct tb_column)*_n);

  for(i=0;i<_n;i++)
  {
    struct tb_column *col=&_t->columns[i];
    col->offset=offset;
    switch(col->type)
    {
    case tb_type_int:
      col->size=sizeof(int);
      break;
    case tb_type_double:
      col->size=sizeof(double);
      break;
    case tb_type_char:
      col->size; //        
      break;
    default:
      assert(false);
    }
    offset+=col->size;
  }

  tb_reallocate(_t,_occurs);
}


//           

int tb_insert(struct tb_table *_t, int _i)
{
  int i;

  //        
  //    (   
  // , )
  assert(_i<=_t->lines_count);

  //        
  tb_reallocate(_t,_t->lines_count+1);

  //      ?
  if(_i==0 && _t->lines_mem<_t->lines)
  { //    
    _t->lines--;
  }else
  {
    //      

    //          1 
    //            
    //  1   

    //     +   ==   
    if(_t->lines-_t->lines_mem+_t->lines_count==_t->lines_count_mem)
    {
      tb_line line;
      _t->lines--;
      line=_t->lines[0];
      memmove(&_t->lines[0],&_t->lines[1],_t->lines_count*sizeof(tb_line));
      _t->lines[_t->lines_count]=line;
    }

    if(_t->lines_count-_i>0)
    {//   
      memmove(&_t->lines[_i+1],&_t->lines[_i],(_t->lines_count-_i)*sizeof(tb_line));
      //     ,     
      _t->lines[_i]=0;
    }
  }

  //    ?
  if(_t->lines[_i]==0)
  {
    //    ,        .
    // tb_column
    int line_length_mem;

    //          
    line_length_mem=_t->columns[_t->columns_count-1].offset+
      _t->columns[_t->columns_count-1].size;

    //     ,     
    _t->lines[_i]=(unsigned char*)malloc(line_length_mem);
  }

  //     
  //   ( int=0, double=0,   *[0]=0
  for(i=0;i<_t->columns_count;i++)
  {
    switch(_t->columns[i].type)
    {
    case tb_type_int:
      *((int*)(_t->lines[_i]+_t->columns[i].offset))=0;
      break;
    case tb_type_double:
      *((double*)(_t->lines[_i]+_t->columns[i].offset))=0;
      break;
    case tb_type_char:
      *(_t->lines[_i]+_t->columns[i].offset)=0;
      break;
    default:
      assert(false);
    }
  }

  _t->lines_count++;

  return _i;
}

int tb_append(struct tb_table *_t)
{
  return tb_insert(_t,_t->lines_count);
}

int tb_getnrows(struct tb_table *_t)
{
  return _t->lines_count;
}

int tb_remove(struct tb_table *_t, int _i)
{
  assert(_i>=0&&_i<_t->lines_count);

  if(_i==0)
  {
    _t->lines++;
  }
  //else
  //{
  //  free(_t->lines[_i]);
  //  memcpy(&_t->lines[_i],&_t->lines[_i+1],(_t->lines_count-(_i+1))*sizeof(tb_line));
  //  //     1 ,    
  //  //    ( memcpy ,    )
  //  //   

  //  _t->lines[_t->lines_count-1]=0;
  //}

  _t->lines_count--;

  if(_t->lines_count==0)
    return -1;

  return _i;
}

int tb_setint(struct tb_table *_t, int _i, int _j, int _v)
{
  assert(_i>=0&&_i<_t->lines_count);
  assert(_j>=0&&_j<_t->columns_count);
  assert(_t->columns[_j].type==tb_type_int);

  *((int*)(_t->lines[_i]+_t->columns[_j].offset))=_v;
  return _v;
}

int tb_getint(struct tb_table *_t, int _i, int _j)
{
  assert(_i>=0&&_i<_t->lines_count);
  assert(_j>=0&&_j<_t->columns_count);
  assert(_t->columns[_j].type==tb_type_int);

  return *((int*)(_t->lines[_i]+_t->columns[_j].offset));
}

double tb_setdouble(struct tb_table *_t, int _i, int _j, double _v)
{
  assert(_i>=0&&_i<_t->lines_count);
  assert(_j>=0&&_j<_t->columns_count);
  assert(_t->columns[_j].type==tb_type_double);

  *((double*)(_t->lines[_i]+_t->columns[_j].offset))=_v;
  return _v;
}

double tb_getdouble(struct tb_table *_t, int _i, int _j)
{
  assert(_i>=0&&_i<_t->lines_count);
  assert(_j>=0&&_j<_t->columns_count);
  assert(_t->columns[_j].type==tb_type_double);

  return *((double*)(_t->lines[_i]+_t->columns[_j].offset));
}

char * tb_setchar(struct tb_table *_t, int _i, int _j, const char *_v)
{
  int len;
  assert(_i>=0&&_i<_t->lines_count);
  assert(_j>=0&&_j<_t->columns_count);
  assert(_t->columns[_j].type==tb_type_char);

  len=(int)strlen(_v)+1;
  assert(_t->columns[_j].size>=len);

  memcpy((char*)(_t->lines[_i]+_t->columns[_j].offset),_v,len);
  return (char*)_v;
}

char * tb_getchar(struct tb_table *_t, int _i, int _j, char *v, int _size)
{
  int len;
  assert(_i>=0&&_i<_t->lines_count);
  assert(_j>=0&&_j<_t->columns_count);
  assert(_t->columns[_j].type==tb_type_char);

  len=(int)strlen((char*)(_t->lines[_i]+_t->columns[_j].offset))+1;
  len=min(_size,len);
  memcpy(v,(char*)(_t->lines[_i]+_t->columns[_j].offset),len);
  v[len-1]=0;
  return v;
}

int tb_findcol(struct tb_table *_t, const char *_name)
{
  int i;
  for(i=0;i<_t->columns_count;i++)
  {
    if(strcmp(_t->columns[i].name,_name)==0)
      return i;
  }
  return -1;
}

void tb_clear(struct tb_table *_t)
{
  _t->lines_count=0;
}

void tb_discard(struct tb_table *_t)
{
  if(_t->lines_mem!=0)
  {
    tb_free_lines(_t);
    free(_t->lines_mem);
    _t->lines_count_mem=0;
    _t->lines_mem=0;
    _t->lines=0;
  }
  if(_t->columns!=0)
  {
    free(_t->columns);
    _t->columns=0;
  }
}

int tb_match(struct tb_table *_t, struct tb_column *_col, int _n, int *_pos, int _exact)
{
  int i,j,count=0;

  for(j=0;j<_n;j++)
  {
    for(i=0;i<_t->columns_count;i++)
    {
      if(_col[j].type==_t->columns[i].type && strcmp(_col[j].name,_t->columns[i].name)==0)
      {
        if(_pos!=0)
          _pos[j]=i;
        count++;
      }
    }
  }

  if(_exact)
    return _t->columns_count==_n && count==_n;
  else
    return count==_n;
}

void tb_getcolumn(struct tb_table *_t, int _j, struct tb_column *_col)
{
  assert(_j>=0&&_j<_t->columns_count);

  memcpy(_col,&_t->columns[_j],sizeof(struct tb_column));
}
